import std/asyncdispatch
import std/os
import std/tables
import std/times
import argparse
import jsony
import nmqtt
import ./lib/asyncsync
import ./lib/config_parser
import ./task/task_di

type
  AppOptions = object
    cfgPath: string
  AppObj = object
    cfg: Config
    mqttctx: MqttCtx
    aq: AsyncQueue[MsgDI]
    tasks: seq[TaskDI]
    di_stats: TableRef[string, int]
  MqttData = object
    timestamp: int64
    state: TableRef[string, int]
  App = ref AppObj

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc newApplication(opts: AppOptions): App =
  result = new App
  result.cfg = parse_config(opts.cfgPath)
  result.aq = newAsyncQueue[MsgDI](32)
  result.di_stats = newTable[string, int]()
  let mqttconf = result.cfg.mqtt
  let mqttctx = newMqttCtx("DI_MQTT")
  mqttctx.set_host(mqttconf.host, mqttconf.port.int)
  mqttctx.set_ping_interval(60)
  result.mqttctx = mqttctx

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc parse_args(): AppOptions =
  let p = newParser("DI-MQTT Emitter"):
    argparse.option("-c", "--config", help = "config file path")
  let opts = p.parse()
  let cfgPath = opts.config
  if opts.help:
    quit(1)
  if not cfgPath.fileExists():
    echo p.help()
    quit(1)
  result.cfgPath = cfgPath

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc launch_tasks(self: App) {.async.} =
  for idx in 0 ..< self.cfg.di.len:
    let param = self.cfg.di[idx]
    let task = newTaskDI(param, self.aq)
    asyncCheck task.run()
    self.tasks.add(task)

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc publish(self: App, timestamp: int64) {.async.} =
  let data = MqttData(timestamp: timestamp, state: self.di_stats)
  let payload = data.toJson()
  await self.mqttctx.publish(self.cfg.mqtt.topic, payload, qos = 0, retain = false)

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc get_ts_ms_now(): int64 =
  result = int64(now().toTime().toUnixFloat() * 1000)

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc mainloop(self: App) {.async.} =
  await self.mqttctx.start()
  await self.launch_tasks()
  var
    fut_msg: Future[MsgDI]
    ts_timeout: int64
    di_msg: MsgDI
    timeouted: bool
    received: bool
  while true:
    timeouted = false
    received = false
    if fut_msg.isNil:
      fut_msg = self.aq.get()
    if ts_timeout == 0:
      # no timeout
      di_msg = await fut_msg
      received = true
    else:
      let ts_now = get_ts_ms_now()
      let wait_ms = ts_timeout - ts_now
      if wait_ms < 0:
        timeouted = true
        received = false
      else:
        received = await withTimeout(fut_msg, wait_ms.int)
        if received:
          di_msg = fut_msg.read()
    if received:
      fut_msg = nil
      self.di_stats[di_msg.name] = di_msg.state
      if ts_timeout == 0 and di_msg.event != DIEvent.Init:
        ts_timeout = get_ts_ms_now() + 1000
        continue
    elif timeouted:
      await self.publish(get_ts_ms_now())
      ts_timeout = 0

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc main() =
  let opts = parse_args()
  let app = newApplication(opts)
  waitFor app.mainloop()

when isMainModule:
  main()
