import std/options
import std/sequtils
import std/strformat
import std/strutils

type
  DIParam* = object
    name*: string   # for MQTT publish
    signal*: string # gpiolib
  MqttConf* = object
    host*: string
    port*: uint16
    topic*: string
  Config* = object
    di*: seq[DIParam]
    mqtt*: MqttConf

type
  Properties = object
    filter: bool

const DI_IDXS = ["S0", "S1", "T0", "T1"]

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc parse_properties(props: string): Properties =
  let parts = props.split(",").mapIt(it.strip).mapIt(it.toLower)
  for prop in parts:
    case prop
    of "raw":
      result.filter = false
    of "filtered":
      result.filter = true
    else:
      discard

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc parse_di(line: string): Option[DIParam] =
  if not line.startsWith("DI_"):
    return
  let parts = line.split("=").mapIt(it.strip)
  if parts.len != 2:
    return
  let keys = parts[0].split("_")
  if keys.len != 2:
    return
  let di_idx = keys[1]
  var props = parts[1].parse_properties()

  var
    signal: string
  let name = &"DI_{di_idx}"
  if di_idx in DI_IDXS:
    signal = &"DIN_{di_idx}"
    props.filter = false
  else:
    var idx_num: int
    try:
      idx_num = di_idx.parseInt()
      if idx_num >= 0 and idx_num < 24:
        let prefix = if props.filter: "Filtered" else: "EXT"
        signal = &"{prefix}_DI{idx_num}"
      else:
        raise newException(ValueError, "Not in range")
    except CatchableError:
      return
  let param = DIParam(name: name, signal: signal)
  result = some(param)

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc parse_config*(path: string): Config =
  var sect = ""
  result.di = newSeqOfCap[DIParam](24)
  # default
  result.mqtt.host = "localhost"
  result.mqtt.port = 1883

  for line in path.readFile.splitLines.mapIt(it.strip).filterIt(it.len > 0):
    if line.startsWith("[") and line.endsWith("]"):
      sect = line[1..^2].toUpper()
      continue
    case sect
    of "DI":
      let param_opt = parse_di(line)
      if param_opt.isSome:
        result.di.add(param_opt.get)
    of "MQTT":
      let parts = line.split("=").mapIt(it.strip)
      if parts.len != 2:
        continue
      case parts[0].toUpper()
      of "HOST":
        result.mqtt.host = parts[1]
      of "PORT":
        try:
          let port = parts[1].parseUInt.uint16
          result.mqtt.port = port
        except CatchableError:
          discard
      of "TOPIC":
        result.mqtt.topic = parts[1]
      else:
        discard

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc `$`*(p: DIParam): string =
  var buf = newSeqOfCap[string](3)
  buf.add(&" - Name:   {p.name}")
  buf.add(&"   Signal: {p.signal}")
  result = buf.join("\n")

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc `$`*(cfg: Config): string =
  var buf = newSeqOfCap[string](32)
  buf.add("=== Config ===")
  buf.add("-- MQTT --")
  buf.add(&"   Host:  {cfg.mqtt.host}:{cfg.mqtt.port}")
  buf.add(&"   Topic: \"{cfg.mqtt.topic}\"")
  buf.add("-- DI --")
  for di in cfg.di:
    buf.add($di)
  result = buf.join("\n")


when isMainModule:
  let cfg = parse_config("../config/di_mqtt.conf")
  echo $cfg
