import std/asyncdispatch
import std/deques
import std/httpclient
import std/json
import std/strformat
import argparse
import nmqtt
import lib/asyncsync
import lib/config_parser
import lib/syslog

type
  MqttTaskObj = object
    cfg: MqttConf
    mqtt: MqttCtx
    aq: AsyncQueue[string]
    event: AsyncEv
    messages: Deque[string]
  MqttTask = ref MqttTaskObj
  AppOptions = object
    cfgPath: string
  AppObj = object
    cfg: MqttConf
    aq: AsyncQUeue[string]
  App = ref AppObj

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc parse_args(): AppOptions =
  let p = newParser("mqtt_soracom"):
    argparse.option("-c", "--config", help = "config file path")
  let opts = p.parse()
  let cfgPath = opts.config
  if opts.help:
    quit(1)
  if not cfgPath.fileExists():
    echo p.help()
    quit(1)
  result.cfgPath = cfgPath

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc newTaskMqtt(cfg: MqttConf, aq: AsyncQueue[string]):
    MqttTask =
  result = new MqttTask
  result.aq = aq
  result.messages = initDeque[string](10)
  result.event = newAsyncEv()
  let mqttctx = newMqttCtx("MQTT_Harvest")
  mqttctx.set_host(cfg.host, cfg.port.int)
  mqttctx.set_ping_interval(60)
  result.cfg = cfg
  result.mqtt = mqttctx

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc wait_mqtt(self: MqttTask) {.async.} =
  while true:
    await self.event.wait()
    self.event.clear()
    while self.messages.len > 0:
      let message = self.messages.popFirst()
      await self.aq.put(message)

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc run(self: MqttTask) {.async.} =
  proc on_data(topic: string, message: string) =
    self.messages.addLast(message)
    self.event.fire()

  await self.mqtt.start()
  await self.mqtt.subscribe(self.cfg.topic, 0, on_data)
  while true:
    await self.wait_mqtt()

# ==============================================================================
#
# ==============================================================================

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc newApplication(cfg: MqttConf): App =
  result = new App
  result.cfg = cfg
  result.aq = newAsyncQueue[string](16)

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc upload(self: App, msg: string) {.async.} =
  let client = newAsyncHttpClient()
  client.headers = newHttpHeaders({"content-type": "application/json"})
  let jsonData = msg.parseJson()
  if jsonData.hasKey("timestamp"):
    let timestamp = jsonData.getOrDefault("timestamp").getInt()
    jsonData.delete("timestamp")
    syslog.info(&"upload: timestamp: {timestamp}")
    client.headers.add("x-soracom-timestamp", $timestamp)
  let body = $(jsonData["state"])
  syslog.info(&"upload: payload: {body}")
  let fut_resp = client.request("http://harvest.soracom.io", httpMethod = HttpPost,
      body = body)
  let res = await withTimeout(fut_resp, 30 * 1000)
  if res:
    let resp = fut_resp.read()
    let code = resp.code
    let logmsg = &"upload: response code: {code}"
    syslog.info(logmsg)
  else:
    syslog.error("upload timeouted.")
  try:
    client.close()
  except CatchableError:
    discard

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc mainloop(self: App) {.async.} =
  while true:
    let msg = await self.aq.get()
    await self.upload(msg)

# ------------------------------------------------------------------------------
#
# ------------------------------------------------------------------------------
proc main(self: App) {.async.} =
  let task_mqtt = newTaskMqtt(self.cfg, self.aq)
  asyncCheck task_mqtt.run()
  await self.mainloop()


when isMainModule:
  syslog.openlog("mqtt_soracom")
  let opts = parse_args()
  let cfg = parse_config(opts.cfgPath)
  let app = newApplication(cfg.mqtt)
  waitFor app.main()
